/*--------------------------------------------
   RealFire with SDL - Roberto Prieto
   Copyright (C) 2001-2012 MegaStorm Systems
  --------------------------------------------
	
	This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.
    
    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
    
    Roberto Prieto
	contact@megastormsystems.com
	http://www.megastormsystems.com

 ------------------------------------------------------------------------
		     RealFire with SDL
	Simple and nice fire effect using SDL for video output. 
	It is optimized in MMX and SSE2 plus standard ASM and C.

	Changelog: Check readme.txt
 ------------------------------------------------------------------------ */

// ---Includes---
#include <windows.h>
#include <time.h>         
#include <string.h>
#include <math.h>
#include "sdl/sdl.h"

#define AlignData(data) (void *)(((int)data + 15) &~ 0x0F) // For aligning data to 16bytes

#define FIRENUM 70
#define XM 800
#define YM 600
#define N_LOOP 10000
#define FIRE_SSE2    // FIRE_NORMAL, FIRE_ASM, FIRE_MMX or FIRE_SSE2
#define ZOOM_SSE2_2    // ZOOM_NORMAL, ZOOM_SSE2-1, ZOOM_SSE2-2 or ZOOM_MMX

// If we defined it, the main loop will be executed during N_LOOP times. For benchmarking
#define BENCH

// Global vars
Uint32 frame_count;
SDL_Surface *screen;
Uint8 *keystate;
// Performance monitor
clock_t iFireTime, iZoomTime, iBlitTime, iTmp;

// Prototypes
void DoRealFire(); 
void MakeFirePalette();

// -DoRealFire()-
void DoRealFire()
{
  Uint32 done;

  // Some vars
  unsigned char *fire_buffer;   
  Uint32 index, fire_num;
  Uint32 dest,fires[FIRENUM];
  Uint32 a,b;
  SDL_Surface *fire800;
  SDL_Color Palette[256];
  Uint32 total_size;
  unsigned char *pixels;

  done=N_LOOP;
  dest=0; 
  index=0; 
  frame_count=0;
  fire_num=FIRENUM;
  iFireTime = iZoomTime = iBlitTime = iTmp = 0;
  for(a=0;a<FIRENUM;a++) fires[a]=5;

  // Fire buffer aligned to 16
  fire_buffer = (unsigned char*)_aligned_malloc(400*301 + 15,16);

  // Fire surface @ 8bits
  fire800 = SDL_CreateRGBSurface(SDL_SWSURFACE,800,600,8,0,0,0,0);
	  
  // Clean surfaces
  SDL_FillRect(screen,NULL,SDL_MapRGB(screen->format,0,0,0));
  SDL_FillRect(fire800,NULL,SDL_MapRGB(fire800->format,100,100,100));
 
  // Create the palette
  for (index=0;index<206;index++)
	{
     Palette[index].r = index+50;
	 Palette[index].g = index;
	 Palette[index].b = 20;
    }	
  for (index=206;index<256;index++)
	{
     Palette[index].r = 0;
	 Palette[index].g = 0;
	 Palette[index].b = 0;
    }	
  SDL_SetColors(fire800, Palette, 0, 256);
  
  // Fire loop
#ifdef BENCH
   while(done>0)
#else
   done=0; 
   while(!done)
#endif
    {
	 for(a=0; a<1200; a++) fire_buffer[dest++]=70;

     for(a=0; a<fire_num; a++)
      {
        if(frame_count>1)
          {
           if(rand()%100>50)
            {
             fires[a]+=rand()&7;
             if(fires[a]>389) fires[a]=(rand()%400)+10;
            }
           else
		   {
            fires[a]-=rand()&7;
            if(fires[a]<11) fires[a]=(rand()%400)+10;
          }
        }
  
      dest=(296*400)+fires[a]-5;
	  for(b=0; b<10; b++, dest++)
        {
         fire_buffer[dest]=rand()&255;
         fire_buffer[dest+400]=rand()&255;
		 fire_buffer[dest+800]=rand()&255;
        }
      }
      
	iTmp = clock();
   // FIRE EFFECT + 2x ZOOM
   // Fire buffer is 400x300 and the zoom will create a final surface of 800x600
   // First 100 lines are not burning at all(can be discarded)
      	 
	total_size=400*197;  // Y=300-3-100

#ifdef FIRE_NORMAL
	// RealFire in C
	unsigned char *myFB;
	int iFB, iT;
	myFB = fire_buffer + 40000;
	for(iFB = 0; iFB < total_size; iFB++)
	{
		iT = (unsigned char)*myFB; 
		iT += (unsigned char)*(myFB + 400); 
		iT += (unsigned char)*(myFB + 800);
		iT += (unsigned char)*(myFB + 800);
		iT += (unsigned char)*(myFB + 1199);
		iT += (unsigned char)*(myFB + 1200);
		iT += (unsigned char)*(myFB + 1200);
		iT += (unsigned char)*(myFB + 1201);
		iT = iT >> 3;
		if(iT > 0) iT--;
		*myFB = iT;
		myFB++;
	}
#endif

#ifdef FIRE_ASM 
	 // RealFire in ASM (non-optimized)
	 _asm {
	       mov edi, fire_buffer
		   add edi, 40000   ; Avoid first 100 lines (do not process them)
		   xor eax, eax
		   xor ebx, ebx
		   mov ecx, total_size
		   top5:		    
		    mov al, byte ptr [edi+1]
			mov bl, byte ptr [edi+401]
		    add eax, ebx
            mov bl, byte ptr [edi+801]
            add eax, ebx
            mov bl, byte ptr [edi+1201]
            add eax, ebx
            mov bl, byte ptr [edi+801]
            add eax, ebx
            mov bl, byte ptr [edi+1201]
            add eax, ebx
            mov bl, byte ptr [edi+1200]
            add eax, ebx
            mov bl, byte ptr [edi+1202]
            add eax, ebx
            shr eax,3
            jz zero4
            dec eax
          zero4:			
            mov byte ptr [edi+1], al
            inc edi			
            dec ecx
			jnz top5		
	 }
#endif

#ifdef FIRE_MMX
	 // RealFire (MMX optimized)
	 Uint64 iSub=0x0001000100010001;
	 _asm {
	       mov edi, fire_buffer
		   add edi, 40000   ; Avoid first 100 lines (do not process them)
		   xor eax, eax
		   xor ebx, ebx
		   mov ecx, total_size
		   pxor mm4,mm4  ; Zero to unpack/pack
		   movq mm5,iSub ; Sub
		   top5:		   
		    ; MMX optimization (8pixels per iteration)
		    movq mm0, qword ptr [edi+1]
			movq mm1, mm0
			; Pack pixels in mm0 and mm1 as 16bit values
			punpcklbw mm0,mm4
			punpckhbw mm1,mm4
			
			; Adding pixels pack as 16bit values and
			movq mm2, qword ptr [edi+401]
			movq mm3, mm2
			punpcklbw mm2,mm4
			punpckhbw mm3,mm4
			paddw mm0,mm2
			paddw mm1,mm3

			movq mm2, qword ptr [edi+801]
			movq mm3, mm2
			punpcklbw mm2,mm4
			punpckhbw mm3,mm4
			paddw mm0,mm2
			paddw mm1,mm3
			paddw mm0,mm2
			paddw mm1,mm3

	        movq mm2, qword ptr [edi+1200]
			movq mm3, mm2
			punpcklbw mm2,mm4
			punpckhbw mm3,mm4
			paddw mm0,mm2
			paddw mm1,mm3

			movq mm2, qword ptr [edi+1201]
			movq mm3, mm2
			punpcklbw mm2,mm4
			punpckhbw mm3,mm4
			paddw mm0,mm2
			paddw mm1,mm3
			paddw mm0,mm2
			paddw mm1,mm3
		
			movq mm2, qword ptr [edi+1202]
			movq mm3, mm2
			punpcklbw mm2,mm4
			punpckhbw mm3,mm4
			paddw mm0,mm2
			paddw mm1,mm3
		   
            ; Shift right to divide the result between 8						
			psraw mm0,3
			psraw mm1,3
			psubusb mm0,mm5
			psubusb mm1,mm5
            
			; Pack pixels in mm0 and store in memory
            packuswb mm0,mm4
			packuswb mm1,mm4			
			psllq mm1,32 ; MMX hasn't rotation, but we can emulate it
			por mm0,mm1			
            movq qword ptr [edi+1], mm0
            
			; Next iteration
			add edi,8
            sub ecx,8
            jnz top5
		  emms ; End MMX instruction		
	 }
#endif

#ifdef FIRE_SSE2
	 // RealFire (SSE2 optimized)
	 Uint64 iSub=0x0001000100010001; // Used for substract 1 per each pixel-packed in 16bits
	 _asm {
	       mov edi, fire_buffer
		   add edi, 40000   ; Avoid first 100 lines
		   xor eax, eax
		   xor ebx, ebx
		   mov ecx, total_size
		   movq xmm4, iSub
		   movq xmm5, iSub
		   pslldq xmm5,8
		   pxor xmm5,xmm4
		   pxor xmm4,xmm4  ; Zero to unpack/pack
		   top5:		   
		    ; SSE2 optimization (16pixels per iteration)
		    movapd xmm0, [edi]
			movapd xmm1, xmm0
			; Unpack 16 pixels on a 128bit SSE register using 16bits for each of them
			punpcklbw xmm0,xmm4
			punpckhbw xmm1,xmm4
			
			; Adding more pixels
			movapd xmm2, [edi+400]
			movapd xmm3, xmm2
			punpcklbw xmm2,xmm4
			punpckhbw xmm3,xmm4
			paddw xmm0,xmm2
			paddw xmm1,xmm3

			; Adding more pixels
			movapd xmm2, [edi+800]
			movapd xmm3, xmm2
			punpcklbw xmm2,xmm4
			punpckhbw xmm3,xmm4
			paddw xmm0,xmm2
			paddw xmm1,xmm3
			paddw xmm0,xmm2
			paddw xmm1,xmm3

			; Adding more pixels
	        movupd xmm2, [edi+1199]
			movapd xmm3, xmm2
			punpcklbw xmm2,xmm4
			punpckhbw xmm3,xmm4
			paddw xmm0,xmm2
			paddw xmm1,xmm3

			; Adding more pixels
			movapd xmm2, [edi+1200]
			movapd xmm3, xmm2
			punpcklbw xmm2,xmm4
			punpckhbw xmm3,xmm4
			paddw xmm0,xmm2
			paddw xmm1,xmm3
			paddw xmm0,xmm2
			paddw xmm1,xmm3
		
			; Adding more pixels
			movupd xmm2, [edi+1201]
			movapd xmm3, xmm2
			punpcklbw xmm2,xmm4
			punpckhbw xmm3,xmm4
			paddw xmm0,xmm2
			paddw xmm1,xmm3
		   
            ; Divide the result between 8 and subtract 1
			psraw xmm0,3
			psraw xmm1,3
			psubusb xmm0,xmm5
			psubusb xmm1,xmm5
            
			; Pack pixels in mm0 and store in memory
            packuswb xmm0,xmm4
			packuswb xmm1,xmm4			
			pslldq xmm1,8 ; Shift 16bits pixel for packing them
			por xmm0,xmm1			
            movapd [edi], xmm0
            
			; Next iteration
			add edi,16
            sub ecx,16
            jnz top5
	 }
#endif

	iFireTime = iFireTime + (clock() - iTmp);
	iTmp = clock();

	pixels=(unsigned char*)fire800->pixels;

#ifdef ZOOM_NORMAL
	// Zoom 2x in C
	{
		unsigned char *myFB,*myFP;
		int iL, iP, iT;
		myFB = fire_buffer + 40000;
		myFP = pixels + 172800;
		for(iL = 0; iL < 192; iL++) // Lines
		{
			for(iP = 0; iP < 400; iP++) // Pixels per line
			{
				iT = (unsigned char)*myFB;
				*(myFP) = iT;
				*(myFP + 1) = iT;
				*(myFP + 800) = iT;
				*(myFP + 801) = iT;
				myFB+=1;
				myFP+=2;
			}
			myFP+=800;
		}
	}
#endif

#ifdef ZOOM_SSE2_1
	// 2x Zoom SSE2 optimized
	unsigned char *aligned;
	aligned = (unsigned char *)AlignData(pixels);

	_asm
	 {
		mov esi, fire_buffer
		add esi, 40000  ;Avoid first 100 lines
		mov edi, aligned
	    add edi, 172800 ;Avoid first 100 lines plus 8 more, multiply per 2 and per 800
		mov ecx, 192    ;300 lines in total but subtract: 100 + 8

		bucle_main21:
			push ecx
			mov ecx, 25    ; 25*32pixels=800pixels(one line)
			
			peque1:
				movapd xmm0, [esi]   ; Move 16 pixels 
				movapd xmm1, xmm0    
				add esi, 16          
				
				punpcklbw xmm0,xmm0  
				punpckhbw xmm1,xmm1  
		
				movapd [edi], xmm0   
				movapd [edi+16], xmm1 
				movapd [edi+800], xmm0
				movapd [edi+816], xmm1
		
		        add edi,32            
				dec ecx                    
				jnz peque1                 

			add edi, 800
			pop ecx
			dec ecx
			jnz bucle_main21
	}
#endif

#ifdef ZOOM_SSE2_2
	// 2x Zoom SSE2 optimized+
	unsigned char *aligned;
	aligned = (unsigned char *)AlignData(pixels);

	_asm
	 {
		mov esi, fire_buffer
		add esi, 40000  ;Avoid first 100 lines
		mov edi, aligned
	    add edi, 172800 ;Avoid first 100 lines plus 8 more, multiply per 2 and per 800
		mov ecx, 192    ;300 lines in total but subtract: 100 + 8

		bucle_main21:
			push ecx
			mov ecx, 5    ; 5*160pixels=800pixels(one line)
			
			peque1:
				movapd xmm0, [esi]            
				movapd xmm1, [esi + 16]
				movapd xmm2, [esi + 32]
				movapd xmm3, [esi + 48]

				movapd xmm4, xmm0
				movapd xmm5, xmm1
				movapd xmm6, xmm2
				movapd xmm7, xmm3
				
				punpcklbw xmm0,xmm0
				punpckhbw xmm4,xmm4
				punpcklbw xmm1,xmm1
				punpckhbw xmm5,xmm5
				punpcklbw xmm2,xmm2
				punpckhbw xmm6,xmm6
				punpcklbw xmm3,xmm3
				punpckhbw xmm7,xmm7
				
				movapd [edi], xmm0             ; Write chain
				movapd [edi+16], xmm4          
				movapd [edi + 32], xmm1  
				movapd [edi + 32 + 16], xmm5   
				movapd [edi + 64], xmm2  
				movapd [edi + 64 + 16], xmm6  
				movapd [edi + 96], xmm3   
				movapd [edi + 96 + 16], xmm7
				movapd [edi + 800], xmm0
				movapd [edi + 816], xmm4
				movapd [edi + 32 + 800], xmm1
				movapd [edi + 32 + 816], xmm5
				movapd [edi + 64 + 800], xmm2
				movapd [edi + 64 + 816], xmm6
				movapd [edi + 96 + 800], xmm3
				movapd [edi + 96 + 816], xmm7

				movapd xmm0, [esi + 64]			; Process the odd pixel
				movapd xmm4, xmm0
				punpcklbw xmm0,xmm0
				punpckhbw xmm4,xmm4
				movapd [edi + 128], xmm0           
				movapd [edi + 128 + 16], xmm4
				movapd [edi + 128 + 800], xmm0
				movapd [edi + 128 + 816], xmm4

				add esi, 80
		        add edi, 160
				dec ecx                    
				jnz peque1                 

			add edi, 800
			pop ecx
			dec ecx
			jnz bucle_main21
	}
#endif

#ifdef ZOOM_MMX
	// 2x Zoom MMX optimized
	_asm
	 {
		mov esi, fire_buffer
		add esi, 40000  ;Avoid first 100 lines)
		mov edi, pixels
	    add edi, 172800 ;Avoid first 100 lines plus 8 more, multiply per 2 and per 800
		mov ecx, 192    ;300 lines in total but subtract: 100 + 8

		bucle_main21:
			push ecx
			mov ecx, 25    ; 25*32pixels=800pixels(one line)
			peque1:
				movq mm0, qword ptr [esi]   ; Move 8 pixels
				movq mm1,mm0                

				movq mm2, qword ptr[esi+8]  
				movq mm3,mm2                

				add esi, 16                 

				punpcklbw mm0,mm0           
				punpckhbw mm1,mm1           
				punpcklbw mm2,mm2
				punpckhbw mm3,mm3

				movq [edi], mm0             ; Write chain
				movq [edi+8], mm1          
				movq [edi+16], mm2
				movq [edi+24], mm3
				movq [edi+800], mm0
				movq [edi+808], mm1
				movq [edi+816],mm2
				movq [edi+824],mm3

		        add edi,32                  
				dec ecx                    
				jnz peque1                 

			add edi, 800
			pop ecx
			dec ecx
			jnz bucle_main21
		emms
	 }
#endif
	   
	 iZoomTime = iZoomTime + (clock() - iTmp);
	 iTmp = clock();

     // Blit to screen. It will automatically convert to screen color depth
	 //SDL_BlitSurface(fire800,NULL,screen,NULL);
	 //SDL_Flip(screen);

	 iBlitTime = iBlitTime + (clock() - iTmp);

#ifdef BENCH
	 done--;
#else
  	 SDL_PumpEvents(); 
	 keystate=SDL_GetKeyState(NULL);
	 if(keystate[SDLK_ESCAPE]) done=1;
#endif
	 frame_count++;
    }

   // Free resources
   //SDL_SaveBMP(fire800,"output.bmp");
   _aligned_free(fire_buffer);
}


int main( int argc, char* argv[] )
{
  clock_t tstart, tend;

  SDL_Init(SDL_INIT_VIDEO);
  //screen=SDL_SetVideoMode(XM,YM,0,SDL_HWSURFACE|SDL_FULLSCREEN);
  screen=SDL_SetVideoMode(XM,YM,0,SDL_SWSURFACE);
  
  tstart=clock();
  DoRealFire();
  tend=clock ();

  printf(" RealFire v1.2\n");
  printf(" ---------------\n\n");
  printf("  - %dx%dx%dbps \n",screen->w,screen->h,screen->format->BitsPerPixel);
  if((screen->flags & SDL_HWSURFACE) == SDL_HWSURFACE) printf("  - Screen on Video Memory (VRAM)\n");
  else printf("  - Screen on System Memory (RAM)\n");
  printf("  - Execution time: %.2f seg\n",(float)(tend-tstart)/(float)CLOCKS_PER_SEC);
  #ifdef BENCH
   printf("  - %2.f frames per second\n", (float)N_LOOP/((float)(tend-tstart)/(float)CLOCKS_PER_SEC));
  #else
   printf("  + %2.f frames per second\n", (float)frame_count/((float)(tend-tstart)/(float)CLOCKS_PER_SEC));
  #endif
   
   // Performance output
   printf("\n\n - DoRealFire() %.2fs - Fire number %d\n",(float)(iFireTime+iZoomTime+iBlitTime)/(float)CLOCKS_PER_SEC,FIRENUM);
   printf("   + Fire: %.2fs - %d%%\n",(float)iFireTime / (float)CLOCKS_PER_SEC, iFireTime * 100 / (iFireTime+iZoomTime+iBlitTime));
   printf("   + Zoom: %.2fs - %d%%\n",(float)iZoomTime / (float)CLOCKS_PER_SEC, iZoomTime * 100 / (iFireTime+iZoomTime+iBlitTime));
   printf("   + Blit: %.2fs - %d%%\n\n\n",(float)iBlitTime / (float)CLOCKS_PER_SEC, iBlitTime * 100 / (iFireTime+iZoomTime+iBlitTime));

  SDL_Quit();
  
  return 0;
}

